import json
import logging
from collections import defaultdict
from pydantic import BaseModel as PydanticBaseModel
from pydantic.fields import Field as PydanticField

from .table import Table
from .utils.get_table_by_name import get_table_by_name


JOIN_SEPARATOR = "____"
logger = logging.getLogger(__name__)


class JoinInfo(PydanticBaseModel):
    model: type
    children: dict[str, "JoinInfo"] = PydanticField(default_factory=dict)

    def add_children(self, path: list[str]):
        name = path[0]
        field = self.model._get_field(name)
        if not field.is_reference:
            raise ValueError(f"Field `{name}` is not a reference (in path `{path}`)")
        if field.reference_type == Table:
            raise ValueError(f"Generic reference cannot be preloaded: {".".join(path)}")
        child = self.children[field.name] = JoinInfo(model=field.reference_type)
        if len(path) > 1:
            child.add_children(path[1:])

    def get_tables_statements(self, parent_alias: str=None):
        if not parent_alias:
            parent_alias = self.model._get_table_name()
            yield f"FROM {parent_alias}"
        for name, child in self.children.items():
            alias = f"{parent_alias}{JOIN_SEPARATOR}{name}"
            yield f"LEFT JOIN {child.model._get_table_name()} AS {alias} ON {alias}.id = {parent_alias}.{name}_id"
            yield from child.get_tables_statements(alias)
    
    def get_columns(self, parent_alias: str=None):
        if not parent_alias:
            parent_alias = self.model._get_table_name()
        for field in self.model._get_fields().values():
            if field.is_reference:
                # scalar reference
                if field.secondary_type is None:
                    if field.base_type == Table:
                        yield f"{parent_alias}{JOIN_SEPARATOR}{field.name}_table", f"{parent_alias}.{field.name}_table"
                    yield f"{parent_alias}{JOIN_SEPARATOR}{field.name}_id", f"{parent_alias}.{field.name}_id"
                # list of references
                elif issubclass(field.base_type, (list, tuple, set)):
                    if field.secondary_type == Table:
                        yield f"{parent_alias}{JOIN_SEPARATOR}{field.name}_tables", f"{parent_alias}.{field.name}_tables"
                    yield f"{parent_alias}{JOIN_SEPARATOR}{field.name}_ids", f"{parent_alias}.{field.name}_ids"
                # ?
                else:
                    raise ValueError()
            else:
                yield f"{parent_alias}{JOIN_SEPARATOR}{field.column_name}", f"{parent_alias}.{field.column_name}"
        # preload via join
        for name, child in self.children.items():
            field = self.model._get_field(name)
            if not field.is_reference:
                continue
            # is this a generic reference?
            alias = f"{parent_alias}{JOIN_SEPARATOR}{name}"
            if field.reference_type != Table:
                yield from child.get_columns(alias)
    
    def get_columns_statements(self):
        for key, value in self.get_columns():
            yield f"{value} AS {key}"

    def get_data(self, row: tuple):
        # fill with data
        def infinite_defaultdict():
            return defaultdict(infinite_defaultdict)
        data = infinite_defaultdict()
        for (alias, _), value in zip(self.get_columns(), row):
            path = alias.split(JOIN_SEPARATOR)[1:]
            item = data
            for p in path[:-1]:
                item = item[p]
            item[path[-1]] = value
        return data
    
    def get_instance(self, row: tuple) -> Table:

        _lazy_joins = {}

        def _get_instance_recursive(data: dict, info: JoinInfo):
            for name, field in info.model._get_fields().items():

                # references...
                if field.is_reference:

                    # scalar reference
                    if field.secondary_type is None:
                        # extract useful data from columns
                        if field.base_type == Table:
                            reference_table = data.pop(f"{name}_table")
                            reference_type = get_table_by_name(reference_table)
                        else:
                            reference_type = field.base_type
                        reference_id = data.pop(f"{name}_id")
                        # format
                        if reference_id is None:
                            data[name] = None
                        elif name in info.children:
                            data[name] = reference_type.load(id=reference_id)
                        else:
                            data.pop(name, None)
                            _lazy_joins[name] = (reference_type, reference_id)

                    # list of references
                    elif issubclass(field.base_type, (list, tuple, set)):
                        # extract useful data from columns
                        references_ids = data.pop(f"{name}_ids")
                        if isinstance(references_ids, str):
                            references_ids = json.loads(references_ids)
                        if field.secondary_type == Table:
                            references_tables = data.pop(f"{name}_tables")
                            if isinstance(references_tables, str):
                                references_tables = json.loads(references_tables)
                            references_types = list(map(get_table_by_name, references_tables))
                        else:
                            references_types = len(references_ids) * [field.secondary_type]
                        # format data
                        if references_ids in (None, []):
                            data[name] = []
                        elif name in info.children:
                            data[name] = [reference_type.load(id=reference_id)
                                          for reference_type, reference_id
                                          in zip(references_types, references_ids)]
                        else:
                            data.pop(name, None)
                            _lazy_joins[name] = (references_types, references_ids)
                    # ?
                    else:
                        raise ValueError()

                # ...other things
                else:
                    data[name] = field.parse(data[name])


            info.model._ensure_lazy_loaders()
            info.model._suspend_validation()
            instance = info.model(**data)
            instance.__dict__.update(data)
            instance._lazy_joins = _lazy_joins
            info.model._resume_validation()
            return instance
        return _get_instance_recursive(self.get_data(row), self)



from typing import Optional

import pytest
from pydantic import Field
from ormantism import Table


def test_specific_foreign_key(setup_db):

    class Node(Table):
        parent: Optional["Node"] = None
        name: str

    grandparent = Node(name="grandparent")
    parent = Node(name="parent", parent=grandparent)
    child = Node(name="child")
    child.parent = parent
    assert grandparent.parent is None
    assert parent.parent.id == grandparent.id
    assert child.parent.id == parent.id

    for preload in ([], ["parent"]):
        grandparent = Node.load(name="grandparent", preload=preload)
        assert grandparent.parent is None
        parent = Node.load(name="parent", preload=preload)
        assert parent.parent.id == grandparent.id
        child = Node.load(name="child", preload=preload)
        assert child.parent.id == parent.id


def test_generic_foreign_key():
    
    class Ref1(Table):
        foo: int
    
    class Ref2(Table):
        bar: int
        
    class Ptr(Table):
        ref: Table

    # creation

    ref1 = Ref1(foo=42)
    ref2 = Ref2(bar=101)
    pointer1 = Ptr(ref=ref1)
    pointer2 = Ptr(ref=ref2)

    # retrieval

    with pytest.raises(ValueError, match="Generic reference cannot be preloaded: ref"):
        pointer1 = Ptr.load(ref=ref1, preload="ref")

    pointer1 = Ptr.load(ref=ref1)
    assert pointer1.ref.id == ref1.id
    assert pointer1.ref.__class__ == Ref1
    pointer2 = Ptr.load(ref=ref2)
    assert pointer2.ref.id == ref2.id
    assert pointer2.ref.__class__ == Ref2

    # update

    pointer2.ref = ref1
    pointer2_id = pointer2.id

    # retrieval

    pointer2 = Ptr.load(id=pointer2_id)
    assert pointer2.ref.id == ref1.id
    assert pointer2.ref.__class__ == Ref1


def test_specific_foreign_key_list(setup_db):

    class Parent(Table):
        name: str
        children: list["Parent"] = Field(default_factory=list)

    n1 = Parent(name="node1")
    n2 = Parent(name="node2")
    n3 = Parent(name="node3", children=[n1, n2])

    assert isinstance(n3.children, list)
    assert len(n3.children) == 2
    assert isinstance(n3.children[0], Parent)

    n3 = Parent.load(name="node3")

    assert isinstance(n3.children, list)
    assert len(n3.children) == 2
    assert isinstance(n3.children[0], Parent)


def test_generic_foreign_key_list():

    class Reference1(Table):
        foo: int

    class Reference2(Table):
        bar: float

    class Reference3(Table):
        foobar: str

    class Pointer(Table):
        ref: list[Table]

    # creation

    reference1 = Reference1(foo=42)
    reference2 = Reference2(bar=3.14)
    reference3 = Reference3(foobar="hello")
    pointer = Pointer(ref=[reference1, reference2])

    # retrieval

    with pytest.raises(ValueError, match="Generic reference cannot be preloaded: ref"):
        pointer = Pointer.load(preload="ref")

    pointer = Pointer.load()
    assert pointer.ref[0].id == reference1.id
    assert pointer.ref[0].__class__ == Reference1
    assert pointer.ref[0].foo == 42
    assert pointer.ref[1].id == reference2.id
    assert pointer.ref[1].__class__ == Reference2
    assert pointer.ref[1].bar == 3.14

    # update

    pointer.ref = [reference3, reference2, reference1]

    # retrieval

    pointer = Pointer.load()
    assert len(pointer.ref) == 3
    assert pointer.ref[0].id == reference3.id
    assert pointer.ref[0].__class__ == Reference3
    assert pointer.ref[0].foobar == "hello"
    assert pointer.ref[1].id == reference2.id
    assert pointer.ref[1].__class__ == Reference2
    assert pointer.ref[1].bar == 3.14
    assert pointer.ref[2].id == reference1.id
    assert pointer.ref[2].__class__ == Reference1
    assert pointer.ref[2].foobar == 42



from typing import ForwardRef
from .get_table_by_name import get_table_by_name


def resolve_type(reference_type: type):
    """Resolve forward reference"""
    if not isinstance(reference_type, ForwardRef):
        return reference_type
    if reference_type.__forward_evaluated__:
        return reference_type.__forward_value__
    cls = get_table_by_name(reference_type.__forward_arg__)
    if cls:
        return cls
    raise ValueError(f"Could not resolve {reference_type}")



from typing import Iterable


def _get_subclasses(base: type) -> Iterable[type]:
    for subclass in base.__subclasses__()[::-1]:
        yield from _get_subclasses(subclass)
        yield subclass


def get_all_tables() -> Iterable[type["Table"]]:
    from ..table import Table
    for cls in _get_subclasses(Table):
        yield cls


def get_table_by_name(name: str) -> type["Table"]:
    for cls in get_all_tables():
        if cls._get_table_name() == name:
            return cls



from __future__ import annotations
import enum
import types
import json
import inspect
import datetime
from typing import Optional, Any, Iterable
from functools import cache
from dataclasses import dataclass, asdict

from pydantic import BaseModel
from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic_core import PydanticUndefined

from .utils.get_base_type import get_base_type
from .utils.resolve_type import resolve_type
from .utils.rebuild_pydantic_model import rebuild_pydantic_model
from .utils.make_hashable import make_hashable
from .utils.supermodel import to_json_schema, from_json_schema


# Define some custom types
# JSON: None | bool | int | float | str | list["JSON"] | dict[str, "JSON"]
JSON = Any


@dataclass
class Field:
    table: type["Table"]
    name: str
    base_type: type
    secondary_type: Optional[type]
    full_type: type
    default: any
    is_required: bool
    column_is_required: bool
    is_reference: bool

    @property
    @cache
    def reference_type(self) -> type:
        if not self.is_reference:
            return None
        if self.secondary_type is None:
            return self.base_type
        return self.secondary_type

    @property
    @cache
    def column_name(self):
        if self.is_reference:
            return f"{self.name}_id"
        return self.name

    @property
    @cache
    def column_base_type(self):
        if self.is_reference:
            return int
        return self.base_type

    @classmethod
    def from_pydantic_info(cls, table: type["Table"], name: str, info: PydanticFieldInfo):
        from .table import Table
        resolved_type = resolve_type(info.annotation)
        base_type, secondary_types, column_is_required = get_base_type(resolved_type)
        secondary_types = [secondary_type for secondary_type in secondary_types if secondary_type != type(None)]
        if len(secondary_types) > 1:
            raise ValueError(f"{table.__name__}.{name}: {secondary_types=}")
        secondary_type = secondary_types[0] if secondary_types else None
        default = None if info.default == PydanticUndefined else info.default
        if info.default_factory:
            default = info.default_factory()
        return cls(table=table,
                   name=name,
                   base_type=base_type,
                   secondary_type=secondary_type,
                   full_type=info.annotation,
                   default=default,
                   column_is_required=column_is_required,
                   is_required=column_is_required and info.is_required(),
                   is_reference=issubclass(base_type, Table) or issubclass(secondary_type, Table))

    @property
    @cache
    def sql_creations(self) -> Iterable[str]:

        # null, default
        sql_null = " NOT NULL" if self.column_is_required else ""
        if self.default is not None:
            serialized = self.serialize(self.default)
            if isinstance(serialized, (int, float)):
                serialized = str(serialized)
            else:
                if not isinstance(serialized, str):
                    serialized = json.dumps(serialized, ensure_ascii=False)
                serialized = "'" + serialized.replace("'", "''") + "'"
            sql_default = f" DEFAULT {serialized}"
        else:
            sql_default = ""

        # references
        if self.is_reference:
            from .table import Table
            # scalar reference
            if self.secondary_type is None:
                if self.base_type == Table:
                    yield f"{self.name}_table TEXT{sql_null}{sql_default}"
                yield f"{self.name}_id INTEGER{sql_null}{sql_default}"
            # list of references
            elif issubclass(self.base_type, (list, tuple, set)):
                if self.secondary_type == Table:
                    yield f"{self.name}_tables JSON{sql_null}{sql_default}"
                yield f"{self.name}_ids JSON{sql_null}{sql_default}"
            return

        # otherwise, only one column to create
        translate_type = {
            bool: "BOOLEAN",
            int: "INTEGER",
            float: "REAL",
            str: "TEXT",
            datetime.datetime: "TIMESTAMP",
            list: "JSON",
            set: "JSON",
            dict: "JSON",
            type[BaseModel]: "JSON",
            type: "JSON",
        }
        if inspect.isclass(self.column_base_type) and issubclass(self.column_base_type, enum.Enum):
            sql = f"{self.column_name} TEXT CHECK({self.column_name} in ('{"', '".join(e.name for e in self.column_base_type)}'))"
        elif inspect.isclass(self.column_base_type) and issubclass(self.column_base_type, BaseModel):
            sql = f"{self.column_name} JSON"
        elif self.column_base_type == JSON:
            sql = f"{self.column_name} JSON DEFAULT 'null'"
        elif self.column_base_type in translate_type:
            sql = f"{self.column_name} {translate_type[self.column_base_type]}"
        else:
            raise TypeError(f"Type `{self.column_base_type}` of `{self.table.__name__}.{self.column_name}` has no known conversion to SQL type")

        # final result
        yield sql + sql_null + sql_default


    def __hash__(self):
        return hash(make_hashable(tuple(asdict(self).items())))

    # conversion

    def serialize(self, value: any):
        if self.base_type == JSON:
            return json.dumps(value, ensure_ascii=False)
        if isinstance(value, bool):
            return int(value)
        if isinstance(value, (int, float, str, type(None))):
            return value
        if isinstance(value, (list, dict)):
            return json.dumps(value, ensure_ascii=False)
        if self.base_type in (set, tuple):
            return json.dumps(list(value))
        if isinstance(value, enum.Enum):
            return value.name
        if self.is_reference:
            return value.id if value else None
        if isinstance(value, types.GenericAlias) or inspect.isclass(value):
            return to_json_schema(value)
        if isinstance(value, BaseModel):
            return value.model_dump_json()
        raise ValueError(f"Cannot serialize value `{value}` of type `{type(value)}` for field `{self.name}`")

    def parse(self, value: any):
        if value is None:
            return None
        if issubclass(self.base_type, enum.Enum):
            return self.base_type[value]
        if self.base_type == JSON:
            if not isinstance(value, str):
                return value
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                return value
        if self.base_type in (dict, list):
            return json.loads(value)
        if self.base_type in (set, tuple):
            return self.base_type(json.loads(value))
        if issubclass(self.base_type, BaseModel):
            return self.base_type(**json.loads(value))
        if self.base_type in (int, float, str, bool):
            return self.base_type(value)
        if self.base_type == datetime.datetime and isinstance(value, str):
            return datetime.datetime.fromisoformat(value)
        if self.base_type == type and not isinstance(value, type):
            if isinstance(value, str):
                value = json.loads(value)
            if not isinstance(value, dict):
                raise ValueError("Type representation should be stored as a `dict`")
            if self.full_type in (type[BaseModel], Optional[type[BaseModel]]):
                return rebuild_pydantic_model(value)
            return from_json_schema(value)
            # raise Exception(value)
        raise ValueError(f"Cannot parse value `{value}` of type `{type(value)}` for field `{self.name}`")
